/*
 * Copyright (c) 2005-2008 jNetX.
 * http://www.jnetx.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * jNetX. You shall not disclose such Confidential Information and
 * shall use it only in accordance with the terms of the license
 * agreement you entered into with jNetX.
 *
 * $Id: IdeaProject.java,v 1.10 2010/02/16 08:02:49 pmoukhataev Exp $
 */
package com.jnetx.ide.intellij;

import com.intellij.ProjectTopics;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.command.CommandProcessor;
import com.intellij.openapi.components.PersistentStateComponent;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.components.State;
import com.intellij.openapi.components.Storage;
import com.intellij.openapi.module.Module;
import com.intellij.openapi.module.ModuleManager;
import com.intellij.openapi.project.ModuleListener;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.roots.JavadocOrderRootType;
import com.intellij.openapi.roots.OrderRootType;
import com.intellij.openapi.roots.libraries.Library;
import com.intellij.openapi.roots.libraries.LibraryTable;
import com.intellij.openapi.roots.libraries.LibraryTablesRegistrar;
import com.intellij.openapi.startup.StartupManager;
import com.intellij.openapi.vfs.JarFileSystem;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.util.xmlb.annotations.AbstractCollection;
import com.jnetx.ide.intellij.util.IdeaStatusBar;
import com.jnetx.ide.intellij.util.VFileUtil;
import org.apache.log4j.Logger;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Represents container that holds set of IDEA's modules
 * (each module used for jNetX component).
 *
 * @author <a href="mailto:pmoukhataev@jnetx.ru">Pavel Moukhataev</a>
 * @version $Revision: 1.10 $
 */
@State(
  name = IdeaProject.JNETX_PROJECT_COMPONENT_NAME,
  storages = {
    @Storage(id="other", file = "$PROJECT_FILE$")
  }
)
public class IdeaProject implements ProjectComponent, ModuleListener, LibraryTable.Listener, PersistentStateComponent<IdeaProject.IdeaProjectSettings> {

    static final Logger log = Logger.getLogger("slee.JnetxProject");

    public static final String LIBRARY_NAME_PREFIX = "jnetx:";
    static final String JNETX_PROJECT_COMPONENT_NAME = "jnetx.project";

    private IdeaStatusBar statusBar;
    private List<String> errors = new ArrayList<String>();

    private Project project;
    private GlobalBuildDir globalBuildDir;
//    private boolean removeLibraries;

    private Map<Module, IdeaModule> ideaModules = new HashMap<Module, IdeaModule>();
    private Map<String, IdeaLibrary> ideaLibraries = new HashMap<String, IdeaLibrary>();
    private Map<String, JnetxComponent> jnetxComponents = new HashMap<String, JnetxComponent>();

    /**
     * Information about closed (and opened) projects to allow later project opening
     *
     * todo: check if this really needed
     */
    private Set<IdeaClosedModule> ideaClosedModules = new HashSet<IdeaClosedModule>();
    private LibraryTable projectLibraryTable;
    private JnetxModuleListLoader jnetxModuleListManager;

    public IdeaProject(Project project) {
        this.project = project;
        jnetxModuleListManager = new JnetxModuleListLoader(this);
        statusBar = new IdeaStatusBar(project);
        globalBuildDir = new GlobalBuildDir(statusBar);
    }

    public Project getProject() {
        return project;
    }

    public IdeaStatusBar getStatusBar() {
        return statusBar;
    }

    public void projectOpened() {
        initGlobalBuildDir();
        jnetxModuleListManager.initJnetxComponentList();

        LibraryTablesRegistrar libraryTablesRegistrar = LibraryTablesRegistrar.getInstance();
        projectLibraryTable = libraryTablesRegistrar.getLibraryTable(project);

        ModuleManager moduleManager = ModuleManager.getInstance(project);
        for (Module module : moduleManager.getModules()) {
            addIdeaModule(module);
        }

        for (Library library : projectLibraryTable.getLibraries()) {
            String libraryName = library.getName();
            if (libraryName != null && libraryName.startsWith(LIBRARY_NAME_PREFIX)) {
                String compName = libraryName.substring(LIBRARY_NAME_PREFIX.length());
                JnetxComponent component = getComponent(compName);
                IdeaLibrary ideaLibrary = new IdeaLibrary(component, library);
                ideaLibraries.put(libraryName, ideaLibrary);
                component.setIdeaLibrary(ideaLibrary);
                log.info("Idea library [" + libraryName + "] recognized as jnetx component : " + compName);
            }
        }
        projectLibraryTable.addListener(this);

        StartupManager.getInstance(project).runWhenProjectIsInitialized(new Runnable() {
            @Override
            public void run() {
                for (IdeaModule module : ideaModules.values()) {
                    module.buildDependencyInfo();
                }
                syncAllDependencies();
            }
        });

        project.getMessageBus().connect(project).subscribe(ProjectTopics.MODULES, this);
    }

    private static final Pattern COMP_FILE_PATTERN = Pattern.compile("^comp-(.*)\\.xml$");

    /**
     * Check if this is jnetx module and if yes add appropriate information.
     * Module is considered as jnetx if it has correct comp-*.xml file.
     * @param module idea module
     * @return IdeaModule or null
     */
    @Nullable
    private IdeaModule addIdeaModule(Module module) {
        VirtualFile moduleFile = module.getModuleFile();
        if (moduleFile == null) return null;
        VirtualFile moduleRoot = moduleFile.getParent();
        if (moduleRoot == null) return null;
        VirtualFile[] moduleFiles = moduleRoot.getChildren();
        List<String> compFileNames = new ArrayList<String>();
        String compName;
        IdeaModule ideaModule = null;
        for (VirtualFile compXmlFile : moduleFiles) {
            Matcher fileNameMatcher = COMP_FILE_PATTERN.matcher(compXmlFile.getName());
            if (!fileNameMatcher.matches()) continue;
            String moduleName = fileNameMatcher.group(1);
            try {
                Element compXml = new SAXBuilder().build(compXmlFile.getInputStream()).getRootElement();
                Element compRootXml = compXml.getChild("component");
                compName = compRootXml.getAttributeValue("name");
                if (!moduleName.equals(compName)) {
                    errors.add("Comp file name doesn't correspond. Module : " + module.getName() + ", Comp file : " + compXmlFile.getName() +  ", comp name : " + compName);
                }
                JnetxComponent jnetxComponent = getComponent(compName);
                ideaModule = new IdeaModule(this, module, jnetxComponent, compXmlFile);
                compFileNames.add(compXmlFile.getName());
                log.info("Idea module [" + module.getName() + "] recognized as jnetx component : " + compName);
            } catch (Exception e) {
                errors.add("XML Parse error : " + e);
                log.error("XML Parse error", e);
            }
        }
        if (compFileNames.size() > 1) {
            errors.add("Several comp files found in module [" + module + "]: " + compFileNames);
        }
        if (ideaModule != null) {
            JnetxComponent jnetxComponent = ideaModule.getJnetxComponent();
            ideaModules.put(module, ideaModule);
            jnetxComponent.setIdeaModule(ideaModule);
        }
        return ideaModule;
    }

    public JnetxModuleListLoader getJnetxModuleListManager() {
        return jnetxModuleListManager;
    }

    private void initGlobalBuildDir() {
        if (globalBuildDir.undefined()) {
            VirtualFile buildDir = project.getBaseDir();
            if (buildDir == null) {
                statusBar.error("Project build dir not defined");
                return;
            }
            if (VFileUtil.isValidDirectory(buildDir)) {
                // assuming that build is located inside of the project dir
                buildDir = buildDir.findChild("build");
            } else {
                statusBar.error("Project dir is not defined");
            }

            globalBuildDir.assignFile(buildDir);
        }
    }

    public JnetxComponent getComponent(String compName) {
        JnetxComponent jnetxComponent = jnetxComponents.get(compName);
        if (jnetxComponent == null) {
            jnetxComponent = new JnetxComponent(compName);
            jnetxComponents.put(compName, jnetxComponent);
        }
        return jnetxComponent;
    }

    public Map<String, JnetxComponent> getJnetxComponents() {
        return jnetxComponents;
    }

    public IdeaModule getIdeaModule(Module module) {
        return ideaModules.get(module);
    }

    public String defineLibraryName(JnetxComponent jnetxComponent) {
        String libraryName = IdeaProject.LIBRARY_NAME_PREFIX + jnetxComponent.getCompName();
        return libraryName;
    }

    private LibraryTable.ModifiableModel libraryModifiableModel;
    private Set<Library.ModifiableModel> createdLibraries;
    public Library createLibrary(JnetxComponent jnetxDependencyComponent) {
        assert jnetxDependencyComponent.getIdeaLibrary() == null : "Library is not present in component";
        if (libraryModifiableModel == null) {
            libraryModifiableModel = projectLibraryTable.getModifiableModel();
            createdLibraries = new HashSet<Library.ModifiableModel>();
        }
        String libraryName = defineLibraryName(jnetxDependencyComponent);
        IdeaLibrary ideaLibrary = ideaLibraries.get(libraryName);
        Library library;
        if (ideaLibrary != null) {
            library = ideaLibrary.getLibrary();
        } else {
            library = this.libraryModifiableModel.createLibrary(libraryName);
            Library.ModifiableModel libraryModifiableModel = library.getModifiableModel();
            updateLibraryContent(jnetxDependencyComponent, libraryName, libraryModifiableModel);
            createdLibraries.add(libraryModifiableModel);
            ideaLibrary = new IdeaLibrary(jnetxDependencyComponent, library);
            ideaLibraries.put(libraryName, ideaLibrary);
        }
        jnetxDependencyComponent.setIdeaLibrary(ideaLibrary);
        return library;
    }

    private void updateLibraryContent(JnetxComponent jnetxDependencyComponent, String libraryName, Library.ModifiableModel libraryModifiableModel) {
        for (String libraryJarURL : libraryModifiableModel.getUrls(OrderRootType.CLASSES)) {
            libraryModifiableModel.removeRoot(libraryJarURL, OrderRootType.CLASSES);
        }
        for (String librarySrcURL : libraryModifiableModel.getUrls(OrderRootType.SOURCES)) {
            libraryModifiableModel.removeRoot(librarySrcURL, OrderRootType.SOURCES);
        }
        for (String libraryJavadocURL : libraryModifiableModel.getUrls(JavadocOrderRootType.getInstance())) {
            libraryModifiableModel.removeRoot(libraryJavadocURL, JavadocOrderRootType.getInstance());
        }
        if (globalBuildDir.undefined()) {
            statusBar.error("Global build dir not found. Unable to create library: " + libraryName + ". <b>Please define global build dir</b>.");
            return;
        }
        addLibraryRootsFromCompXml(jnetxDependencyComponent, libraryModifiableModel);
        // libraryModifiableModel.addJarDirectory(jnetxDependencyComponent.root.getAbsoluteFile().toURI().toString(), true);

    }

    private void addLibraryRootsFromCompXml(JnetxComponent jnetxDependencyComponent, Library.ModifiableModel libraryModifiableModel) {
        for (VirtualFile repositoryDir : globalBuildDir.getRepositories()) {
            VirtualFile compDir = repositoryDir.findChild(jnetxDependencyComponent.getCompName());
            if (VFileUtil.isValidDirectory(compDir)) {
                addProjectLibraryFiles(compDir, libraryModifiableModel);
                // Parse comp.xml and add idea
                try {
                    VirtualFile compXmlFile = compDir.findChild("comp-" + jnetxDependencyComponent.getCompName() + ".xml");
                        Element componentsXml = new SAXBuilder().build(compXmlFile.getInputStream()).getRootElement();
                        Element componentXml = componentsXml.getChild("component");
                        Element compIdeXml = componentXml.getChild("ide");
                        if (compIdeXml != null) {
                            addLibraryRootFromCompXml(libraryModifiableModel, compDir, compIdeXml, "source", OrderRootType.SOURCES);
                            addLibraryRootFromCompXml(libraryModifiableModel, compDir, compIdeXml, "javadoc", JavadocOrderRootType.getInstance());
                        }
                } catch (NullPointerException e) {
                    log.error("Can't parse comp.xml", e);
                } catch (Exception e) {
                    log.error("Exception occured", e);
                }
            }
        }
    }

    @SuppressWarnings({"unchecked"})
    private void addLibraryRootFromCompXml(Library.ModifiableModel libraryModifiableModel, VirtualFile compDir
            , Element compIdeXml, String tagName, OrderRootType rootType) {
        for (Element sourceXml : (List<Element>) compIdeXml.getChildren(tagName)) {
            String urlString = sourceXml.getAttributeValue("url");
            if (urlString == null) {
                log.warn("IDE element doesn't contain URL : " + compDir + ", " + tagName);
            }
            int protocolLen = urlString.indexOf(':');
            if (protocolLen > 0 && urlString.length() > protocolLen + 2) {
                if (urlString.charAt(protocolLen + 1) != '/' && urlString.charAt(protocolLen + 2) != '/') {
                    boolean additionalDelimiterNeeded = urlString.charAt(protocolLen + 1) != '/';
                    urlString = urlString.substring(0, protocolLen + 1)
                            + "//" + compDir.getPath()
                            + (additionalDelimiterNeeded ? "/" : "")
                            + urlString.substring(protocolLen + 1);
                }
                libraryModifiableModel.addRoot(urlString, rootType);
            } else {
                log.error("Unknown protocol : " + urlString);
            }
        }
    }

    private void addProjectLibraryFiles(VirtualFile file, Library.ModifiableModel libraryModifiableModel) {
        if (VFileUtil.isInvalidFile(file)) {
            errors.add("Wrong file");
        } else if (file.isDirectory()) {
            for (VirtualFile file1 : file.getChildren()) {
                addProjectLibraryFiles(file1, libraryModifiableModel);
            }
        } else if (file.exists()) {
            if (file.getName().endsWith(".jar")) {
                JarFileSystem jarFileSystem = JarFileSystem.getInstance();
                VirtualFile jarFile = jarFileSystem.getJarRootForLocalFile(file);
                if (jarFile != null) {
                    libraryModifiableModel.addRoot(jarFile, OrderRootType.CLASSES);
                }
            }
        }
    }

    public void commitLibraryModifiableModel() {
        if (libraryModifiableModel != null) {
            for (Library.ModifiableModel createdLibrary : createdLibraries) {
                log.info("New library created:" + createdLibrary);
                createdLibrary.commit();
            }
            libraryModifiableModel.commit();
            libraryModifiableModel = null;
            createdLibraries = null;
        }
    }

    public void projectClosed() {
        for (IdeaModule moduleManager : ideaModules.values()) {
            moduleManager.close();
        }
        ideaModules.clear();
    }

    @NotNull
    public String getComponentName() {
        return JNETX_PROJECT_COMPONENT_NAME;
    }

    public void initComponent() {

    }

    public void disposeComponent() {

    }

    @Override
    public IdeaProjectSettings getState() {
        return new IdeaProjectSettings(globalBuildDir.getPath(), ideaClosedModules);
    }

    @Override
    public void loadState(IdeaProjectSettings state) {
        if (state.getGlobalBuildDir() != null && !state.getGlobalBuildDir().equals("null")) {
            globalBuildDir.assignPath(state.getGlobalBuildDir());
        }
        ideaClosedModules = state.getIdeaClosedModules();
    }

    public void syncAllDependencies() {
        for (IdeaModule ideaModule : ideaModules.values()) {
            ideaModule.syncAndCommit();
        }
    }

    public GlobalBuildDir getGlobalBuildDir() {
        return globalBuildDir;
    }

    public void setGlobalBuildDir(VirtualFile buildDir) {
        if (globalBuildDir.assignFile(buildDir)) {
            // if global build dir has changed
            ApplicationManager.getApplication().runWriteAction(
                    new Runnable() {
                        @Override
                        public void run() {
                            updateLibraries();
                        }
                    });
        }
    }

    public void updateLibraries() {
        LibraryTable.ModifiableModel libraryTableModifiableModel = null;
        for (IdeaLibrary ideaLibrary : ideaLibraries.values()) {
            JnetxComponent jnetxDependencyComponent = ideaLibrary.getJnetxComponent();
            String libraryName = IdeaProject.LIBRARY_NAME_PREFIX + jnetxDependencyComponent.getCompName();
            Library library = projectLibraryTable.getLibraryByName(libraryName);
            if (library == null) {
                if (libraryTableModifiableModel == null) {
                    libraryTableModifiableModel = projectLibraryTable.getModifiableModel();
                }
                library = libraryTableModifiableModel.createLibrary(libraryName);
                ideaLibrary.setLibrary(library);
            }
            Library.ModifiableModel libraryModifiableModel = library.getModifiableModel();
            updateLibraryContent(jnetxDependencyComponent, jnetxDependencyComponent.getCompName(), libraryModifiableModel);
            libraryModifiableModel.commit();
        }

        if (libraryTableModifiableModel != null) {
            libraryTableModifiableModel.commit();
        }
    }

    public Set<IdeaClosedModule> getIdeaClosedModules() {
        return ideaClosedModules;
    }

    public void moduleAdded(Project project, final Module module) {
        CommandProcessor.getInstance().executeCommand(project, new Runnable() {
            @Override
            public void run() {
                ApplicationManager.getApplication().runWriteAction(new Runnable() {
                    public void run() {
                        IdeaModule ideaModule = addIdeaModule(module);
                        if (ideaModule != null) {
                            ideaModule.buildDependencyInfo();
                        }
                        syncAllDependencies();
                    }
                });
            }
        }, "Add module " + module, null);
    }

    public void beforeModuleRemoved(Project project, Module module) {

    }

    public void moduleRemoved(Project project, final Module module) {
        CommandProcessor.getInstance().executeCommand(project, new Runnable() {
            @Override
            public void run() {
                ApplicationManager.getApplication().runWriteAction(new Runnable() {
                    public void run() {
                        IdeaModule ideaModule = ideaModules.remove(module);
                        if (ideaModule != null) {
                            ideaModule.getJnetxComponent().setIdeaModule(null);
                        }
                        String moduleName = module.getName();
                        ideaClosedModules.add(new IdeaClosedModule(moduleName, module.getModuleFilePath()));

                        syncAllDependencies();
                    }
                });
            }
        }, "Remove module " + module, null);
    }

    public void modulesRenamed(Project project, List<Module> modules) {
// todo [!] implement this       
//        String moduleName = module.getName();
//        if (ideaClosedModules.containsKey(moduleName)) return;
//        ideaClosedModules.put(moduleName, new JnetxModuleInfo(moduleName, module.getModuleFilePath()));
    }

    @Override
    public void afterLibraryAdded(Library newLibrary) {

    }

    @Override
    public void afterLibraryRenamed(Library library) {

    }

    @Override
    public void beforeLibraryRemoved(Library library) {

    }

    /**
     * TODO [!] on library remove some synchronization must be done
     * @param library library
     */
    @Override
    public void afterLibraryRemoved(Library library) {
        String libraryName = library.getName();
        ideaLibraries.remove(libraryName);
        if (libraryName != null && libraryName.startsWith(LIBRARY_NAME_PREFIX)) {
            String compName = libraryName.substring(LIBRARY_NAME_PREFIX.length());
            JnetxComponent component = getComponent(compName);
            component.setIdeaLibrary(null);
            log.info("Removed Idea library [" + libraryName + "] recognized as jnetx component : " + compName);
        } else {
            log.info("Removed Idea library [" + libraryName + "] not recognized as jnetx component.");
        }
    }

    /**
     * Information about all ide modules/jnetx components (opened and closed)
     */
    @SuppressWarnings({"UnusedDeclaration"})
    public static final class IdeaClosedModule {
        public String moduleName;
        public String modulePath;

        public IdeaClosedModule() {
        }

        public IdeaClosedModule(String moduleName, String modulePath) {
            this.moduleName = moduleName;
            this.modulePath = modulePath;
        }

        public String getModuleName() {
            return moduleName;
        }

        public String getModulePath() {
            return modulePath;
        }
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public static final class IdeaProjectSettings {
        public String globalBuildDir;
        @AbstractCollection(elementTypes = IdeaClosedModule.class, elementTag = "closedModule", surroundWithTag = true)
        public Set<IdeaClosedModule> ideaClosedModules;

        public IdeaProjectSettings() {
            ideaClosedModules = new HashSet<IdeaClosedModule>();
        }

        public IdeaProjectSettings(String globalBuildDir, Set<IdeaClosedModule> ideaClosedModules) {
            this.globalBuildDir = globalBuildDir;
            this.ideaClosedModules = ideaClosedModules;
        }

        public String getGlobalBuildDir() {
            return globalBuildDir;
        }

        public Set<IdeaClosedModule> getIdeaClosedModules() {
            return ideaClosedModules;
        }
    }
}
